/* -*- c++ -*-
 * tkresolver - Telekinesis testing software
 *
 * Copyright (C) 2008 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <memory>

#include "libtkquery/parse.h"
#include "libtkquery/parse_error.h"

#include "editor/constraint.h"

using namespace Tk::Editor;

Constraint:: Constraint(Tk::Db::Database *database, Domain &domain)
  : _database(database),
    _domain(domain.get_domain()),
    _valid(false)
{
  WDEBUG_OUT << L"initialising Constraint" << std::endl;
  evaluate();
  domain.signal_domain_changed().connect( sigc::mem_fun(*this, &Constraint::set_domain) );
  WDEBUG_OUT << L"initialised Constraint" << std::endl;
}

void
Constraint:: set_valid(bool valid, const Glib::ustring &tooltip)
{
  _valid = valid;
  
  if (valid)
    {
      set_has_tooltip(false);
      modify_base(Gtk::STATE_NORMAL, Gdk::Color("pale green"));
    }
  else
    {
      if (tooltip.size() > 0)
	set_tooltip_text(static_cast<std::string>(tooltip));
      else
	set_has_tooltip(false);
      
      modify_base(Gtk::STATE_NORMAL, Gdk::Color("tomato"));
    }
}

Tk::Query::Query *
Constraint:: get_query(const Tk::String &query_string)
{
  Tk::Query::ExpressionPtr expr( Tk::Query::parse_string(_domain, query_string) );
  return expr->build_query(_database, _domain);
}

void
Constraint:: on_changed()
{
  evaluate();
}

void
Constraint:: evaluate()
{
  //WDEBUG_OUT << L"Evaluating" << std::endl;
  Tk::String query_str( get_text() );
  if (get_text().size() < 1)
    {
      set_valid(true);
      return;
    }
  
  try
    {
      Tk::Query::QueryPtr query( get_query(query_str) );
      set_valid(true);
    }
  catch (Tk::Query::ParseError &e)
    {
      set_valid(false,
		e.get_type()
		+ " error at character "
		+ static_cast<std::string>(Tk::String(e.get_location().begin.column))
		+ ": "
		+ e.get_message()
		);
    }
}

void
Constraint:: on_activate()
{
  if (!_valid)
    return;
  
  if (get_text().length() > 0)
    {
      std::auto_ptr<Tk::Query::Query> query(get_query());
      _signal_constraint_activated.emit(query->get_entities(_domain));
    }
  else
    {
      _signal_constraint_activated.emit(Tk::Query::get_all_entities(_database, _domain));
    }
}

Constraint::SignalConstraintActivated
Constraint:: signal_constraint_activated()
{
  return _signal_constraint_activated;
}

void
Constraint:: set_domain(const Tk::String &domain)
{
  _domain = domain;
  evaluate();
}

Tk::Query::Query *
Constraint:: get_query()
{
  Glib::ustring text(get_text());
  if (!_valid || text.length() < 1)
    return 0;
  
  Tk::Query::ExpressionPtr expr( Tk::Query::parse_string(_domain, text) );
  return expr->build_query(_database, _domain);
}
